/***********************************
 Bridging Graphics:
 provides functionality for simple graphics

 Copyright (c) 2006-2007, Lori Scarlatos

 Module Name:
	bg.cpp

 Module Purpose:
	define BG functions
***********************************/

#include "bg.h"

#include <stdio.h>
#include <math.h>
#include <malloc.h>
#include <string.h>
#include <stdlib.h>

/***** GLOBAL VARIABLES *****/

// Global variables used in drawing graphics (private)

// Named colors

GLfloat bgCLUT[13][3] = {
        {0.0, 0.0, 0.0}, // BG_BLACK
        {1.0, 1.0, 1.0}, // BG_WHITE
        {0.5, 0.5, 0.5}, // BG_GREY
        {1.0, 0.0, 0.0}, // BG_RED
        {0.0, 1.0, 0.0}, // BG_GREEN
        {0.0, 0.0, 1.0}, // BG_BLUE
        {1.0, 1.0, 0.0}, // BG_YELLOW
        {0.0, 1.0, 1.0}, // BG_CYAN
        {1.0, 0.0, 1.0}, // BG_MAGENTA
        {1.0, 0.5, 0.0}, // BG_ORANGE
        {1.0, 0.5, 0.5}, // BG_PINK
        {0.4, 0.2, 0.0}, // BG_BROWN
        {0.6, 0.2, 0.8}, // BG_PURPLE
};

// Color arrays, used to draw Hue values

GLfloat bgRedHueArray[360];
GLfloat bgGreenHueArray[360];
GLfloat bgBlueHueArray[360];

GLfloat bgRedHueMap[256];
GLfloat bgGreenHueMap[256];
GLfloat bgBlueHueMap[256];

GLfloat bgRedColorMap[256];
GLfloat bgGreenColorMap[256];
GLfloat bgBlueColorMap[256];

// Current location, used in drawing

GLfloat bgLocation[2] = {0.0, 0.0};

// Sine and cosine tables used for drawing circles

GLfloat bgSineTable[257];
GLfloat bgCosineTable[257];

/*
// Window constants
HWND hWnd;
HDC hDC;
HGLRC hRC;
*/

static GLint bgWindW, bgWindH;	/* size of the window on the screen */

static float bgWidth, bgHeight; /* number of units in coordinate system */
static float bgWorldRatio;		/* ratio of height / width */
static float bgLeft;
static float bgRight;
static float bgBottom;
static float bgTop;

/***** LOCAL FUNCTIONS (NOT CALLED BY THE USER) *****/

#define PI 3.1415926

// Initialize global variables

void bgInit (void) {
   float angle, angle_increment;
   float sine_value;
   int i1, i2, i3, i4, j1, j2, j3, j4;
   int h;
   float f, f_increment;

   // Initialize sine/cosine table

   angle_increment = PI / 128.0;
   for (i1=0, i2=128, i3=128, i4=256, j1=192, j2=64, j3=64, j4=192, angle = 0.0;
        i1 <= 64;
        i1++, i2--, i3++, i4--, j1++, j2--, j3++, j4--, angle += angle_increment) {
          sine_value = sin(angle);
          bgSineTable[i1] = bgSineTable[i2] = sine_value;
          bgSineTable[i3] = bgSineTable[i4] = -sine_value;
          bgCosineTable[j1] = bgCosineTable[j2] = sine_value;
          bgCosineTable[j3] = bgCosineTable[j4] = -sine_value;
   }

   // Initialize hue color array:
   // assume full saturation and value for 360 colors on color wheel
   f_increment = 1.0/60.0;
   for (h=0, f=0.0; h<60; h++, f += f_increment) {  // Hue / 60 = 0
       bgRedHueArray[h] = 1.0;
       bgGreenHueArray[h] = f;
       bgBlueHueArray[h] = 0.0;
   }
   for (h=60, f=1.0; h<120; h++, f -= f_increment) {  // Hue / 60 = 1
       bgRedHueArray[h] = f;
       bgGreenHueArray[h] = 1.0;
       bgBlueHueArray[h] = 0.0;
   }
   for (h=120, f=0.0; h<180; h++, f += f_increment) {  // Hue / 60 = 2
       bgRedHueArray[h] = 0.0;
       bgGreenHueArray[h] = 1.0;
       bgBlueHueArray[h] = f;
   }
   for (h=180, f=1.0; h<240; h++, f -= f_increment) {  // Hue / 60 = 3
       bgRedHueArray[h] = 0.0;
       bgGreenHueArray[h] = f;
       bgBlueHueArray[h] = 1.0;
   }
   for (h=240, f=0.0; h<300; h++, f += f_increment) {  // Hue / 60 = 4
       bgRedHueArray[h] = f;
       bgGreenHueArray[h] = 0.0;
       bgBlueHueArray[h] = 1.0;
   }
   for (h=300, f=1.0; h<360; h++, f -= f_increment) {  // Hue / 60 = 5
       bgRedHueArray[h] = 1.0;
       bgGreenHueArray[h] = 0.0;
       bgBlueHueArray[h] = f;
   }

	//Initialize color maps using CLUT
   for (int i = 0; i < 13; i++) {
       bgRedColorMap[i] = bgCLUT[i][0];
       bgGreenColorMap[i] = bgCLUT[i][1];
       bgBlueColorMap[i] = bgCLUT[i][2];
   }
       
	/* Create color maps for hue images */
   for (i1 = i2 = 0; i2 < 360; i1++, i2+=2) {
       bgRedHueMap[i1] = bgRedHueArray[i2];
       bgGreenHueMap[i1] = bgGreenHueArray[i2];
       bgBlueHueMap[i1] = bgBlueHueArray[i2];
   }

   // Set sizes for lines and points
   glPointSize(5.0);
   glLineWidth(3.0);

   // Set blending functions for images that are partially transparent
   glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
}

/* Window Procedure

LRESULT CALLBACK WndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
   switch ( message ) {

     case WM_CREATE:
        return 0;

     case WM_CLOSE:
        PostQuitMessage( 0 );
        return 0;

     case WM_DESTROY:
        return 0;

     case WM_KEYDOWN:
        switch ( wParam ) {

          case VK_ESCAPE:
               PostQuitMessage( 0 );
               return 0;
        }
        return 0;

     default:
        return DefWindowProc( hWnd, message, wParam, lParam );
   }
}

// Enable OpenGL

VOID EnableOpenGL( HWND hWnd, HDC * hDC, HGLRC * hRC )
{
     PIXELFORMATDESCRIPTOR pfd;
     int iFormat;

     // get the device context (DC)
     *hDC = GetDC( hWnd );

     // set the pixel format for the DC
     ZeroMemory( &pfd, sizeof( pfd ) );
     pfd.nSize = sizeof( pfd );
     pfd.nVersion = 1;
     pfd.dwFlags = PFD_DRAW_TO_WINDOW |
     PFD_SUPPORT_OPENGL | PFD_DOUBLEBUFFER;
     pfd.iPixelType = PFD_TYPE_RGBA;
     pfd.cColorBits = 24;
     pfd.cDepthBits = 16;
     pfd.iLayerType = PFD_MAIN_PLANE;
     iFormat = ChoosePixelFormat( *hDC, &pfd );
     SetPixelFormat( *hDC, iFormat, &pfd );

     // create and enable the render context (RC)
     *hRC = wglCreateContext( *hDC );
     wglMakeCurrent( *hDC, *hRC );

     // Set range of coordinate values
     glMatrixMode(GL_PROJECTION);
     glLoadIdentity();
     glOrtho(bgLeft, bgRight, bgBottom, bgTop, -10.0, 10.0); // Range of coordinates in the viewport
     glMatrixMode(GL_MODELVIEW);
}

// Disable OpenGL

VOID DisableOpenGL( HWND hWnd, HDC hDC, HGLRC hRC )
{
     wglMakeCurrent( NULL, NULL );
     wglDeleteContext( hRC );
     ReleaseDC( hWnd, hDC );
} 
*/

// Hidden callbacks

/* Draw the picture on the screen */

#define BG_DISPLAYLIST 1

void Draw(void) {
	// Draw the display list
	glCallList(BG_DISPLAYLIST);

	// Update the screen
    glFlush();			/* force all changes to be applied to buffer */
	glutSwapBuffers();	/* display what you just drew */
}

/* Resize the picture */

void Resize(GLint w, GLint h) 
{
	float windowRatio;

	// Check the new ratios
	windowRatio = (float)h / w;

	if (bgWorldRatio < windowRatio) {
		bgWindW = w;
		bgWindH = (bgHeight * w) / bgWidth;
	}
	else if (bgWorldRatio > windowRatio) {
		bgWindH = h;
		bgWindW = (bgWidth * h) / bgHeight;
	}
	else {
		bgWindW = w;
		bgWindH = h;
	}

    glViewport(0, 0, bgWindW, bgWindH);
}

/***** PUBLIC FUNCTIONS (DEFINED IN BG.H) *****/
/***** Window functions *****/

void bgCreateWindow (void)
{
	int argc = 1;
	char *argv = "main.exe";

	// Set initial window dimensions
	bgWindW = bgWindH = 512;

	bgWidth = 32.0;
	bgHeight = 32.0;
	bgWorldRatio = 1.0;
	bgLeft = -16.0;
	bgRight = 16.0;
	bgBottom = -16.0;
	bgTop = 16.0;

	// Initialize the GLUT window & open it
	glutInit(&argc, &argv);
	glutInitWindowPosition(0, 0); 
	glutInitWindowSize(bgWindW, bgWindH);
	glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE);

    if (glutCreateWindow("Bridging Graphics Application") == GL_FALSE) {
		exit(1);
    }

     // Set range of coordinate values
     glMatrixMode(GL_PROJECTION);
     glLoadIdentity();
     glOrtho(bgLeft, bgRight, bgBottom, bgTop, -10.0, 10.0); // Range of coordinates in the viewport
     glMatrixMode(GL_MODELVIEW);

	// Set default display & resize callbacks
    glutReshapeFunc(Resize);
    glutDisplayFunc(Draw);

	// Finish BG initialization
	bgInit();
}

void bgCreateWindow (enum BGwindowtypes w)
{
	int argc = 1;
	char *argv = "main.exe";

	// Set initial window dimensions
	switch (w) {
	case BG_RECTANGLE :
		bgWindW = 640;
		bgWindH = 480;

		bgWidth = 40.0;
		bgHeight = 30.0;
		bgWorldRatio = 0.75;
		bgLeft = 0.0;
		bgRight = 40.0;
		bgBottom = 0.0;
		bgTop = 30.0;
		break;
	case BG_FULL :
	default :
		bgWindW = bgWindH = 512;

		bgWidth = 32.0;
		bgHeight = 32.0;
		bgWorldRatio = 1.0;
		bgLeft = -16.0;
		bgRight = 16.0;
		bgBottom = -16.0;
		bgTop = 16.0;
	}
	// Initialize the GLUT window & open it
	glutInit(&argc, &argv);
	glutInitWindowPosition(0, 0); 
	glutInitWindowSize(bgWindW, bgWindH);
	glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE);

    if (glutCreateWindow("Bridging Graphics Application") == GL_FALSE) {
		exit(1);
    }

     // Set range of coordinate values
     glMatrixMode(GL_PROJECTION);
     glLoadIdentity();
     glOrtho(bgLeft, bgRight, bgBottom, bgTop, -10.0, 10.0); // Range of coordinates in the viewport
     glMatrixMode(GL_MODELVIEW);

	// Set default display & resize callbacks
    glutReshapeFunc(Resize);
    glutDisplayFunc(Draw);

	// Finish BG initialization
	bgInit();
}

extern float bgGetWindowWidth (void)
{
	return bgWidth;
}

extern float bgGetWindowHeight (void)
{
	return bgHeight;
}

extern float bgGetWindowLeft (void)
{
	return bgLeft;
}

extern float bgGetWindowRight (void)
{
	return bgRight;
}

extern float bgGetWindowBottom (void)
{
	return bgBottom;
}

extern float bgGetWindowTop (void)
{
	return bgTop;
}

// Loop until window is closed

void bgMainLoop (void)
{
    glutMainLoop();
}

// wait a specified number of seconds
void bgPause (float seconds)
{
    // NOTE: This only works in Windows!
	Sleep (int(seconds*1000));
}     

/***** Set Colors *****/

// Set the color using constant values indexing bgCLUT

void bgSetColor (enum BGcolors C)
{
     glColor3fv(bgCLUT[C]);
}

// Set the background color using constant values indexing bgCLUT

void bgSetBackgroundColor (enum BGcolors C)
{
     glClearColor(bgCLUT[C][0], bgCLUT[C][1], bgCLUT[C][2], 0);
}

// Set the color

void bgSetHue (int h)
{
     // Check values to ensure they are in range
     if (h < 0) h = 0;
     else if (h > 360) h = h % 360;

     glColor3f(bgRedHueArray[h], bgGreenHueArray[h], bgBlueHueArray[h]);
}

// Set the color

void bgSetGrey (float grey_value)
{
     // Check values to ensure they are in range
     if (grey_value < 0) grey_value = 0;
     else if (grey_value > 1) grey_value = 1.0;

     glColor3f(grey_value, grey_value, grey_value);
}

// Set the color

void bgSetBackgroundRGB (float R, float G, float B)
{
     // Check values to ensure they are in range
     if (R < 0) R = 0;
     else if (R > 1) R = 1.0;
     if (G < 0) G = 0;
     else if (G > 1) G = 1.0;
     if (B < 0) B = 0;
     else if (B > 1) B = 1.0;

     glClearColor(R, G, B, 0);
}

// Set the color using RGB values from 0..1

void bgSetRGB (float R, float G, float B)
{
     // Check values to ensure they are in range
     if (R < 0) R = 0;
     else if (R > 1) R = 1.0;
     if (G < 0) G = 0;
     else if (G > 1) G = 1.0;
     if (B < 0) B = 0;
     else if (B > 1) B = 1.0;

     glColor3f(R, G, B);
}


// Set the color using RGB values from 0..255
void bgSetRGBbyte (unsigned char R, unsigned char G, unsigned char B)
{
     glColor3ub(R, G, B);
}

/***** Drawing Functions *****/

// Clear the screen with the background color

void bgClearScreen (void)
{
	// First, start a new Display List
	glNewList(BG_DISPLAYLIST, GL_COMPILE);

	// Now, clear the screen
    glClear(GL_COLOR_BUFFER_BIT);
}

// Change the current location

void bgMoveTo (float x, float y)
{
     bgLocation[0] = x;
     bgLocation[1] = y;
}

// Draw from current location to new point, then update current location

void bgDrawTo (float x, float y)
{
     // Draw the line
     glBegin(GL_LINES);
         glVertex2fv(bgLocation);
         glVertex2f(x, y);
     glEnd();

     // Update the location
     bgLocation[0] = x;
     bgLocation[1] = y;
}

void bgDrawPoint (float x, float y)
{
     // Update the location
     bgLocation[0] = x;
     bgLocation[1] = y;

     // Draw the point
     glBegin(GL_POINTS);
         glVertex2fv(bgLocation);
     glEnd();

}

void bgDrawBox (float x1, float y1, float x2, float y2)
{
     glRectf(x1, y1, x2, y2);
}

void bgDrawTriangle (float x1, float y1, float x2, float y2, float x3, float y3)
{
     glBegin(GL_TRIANGLES);
         glVertex2f(x1, y1);
         glVertex2f(x2, y2);
         glVertex2f(x3, y3);
     glEnd();
}

void bgDrawCircle (float x1, float y1, float radius)
{
     int numOfIterations, stepSize;
     int i, j;

     // Calculate how many line segments go around the circle,
     // and how to step through the Sine and Cosine tables
     if (radius >= 16) {
        numOfIterations = 256;
        stepSize = 1;
     }
     else if (radius >= 8) {
        numOfIterations = 128;
        stepSize = 2;
     }
     else if (radius >= 4) {
        numOfIterations = 64;
        stepSize = 4;
     }
     else if (radius >= 2) {
        numOfIterations = 32;
        stepSize = 8;
     }
     else if (radius >= 1) {
        numOfIterations = 16;
        stepSize = 16;
     }
     else {
        numOfIterations = 8;
        stepSize = 32;
     }

     glPushMatrix();
     glTranslatef(x1, y1, 0.0);

     glBegin(GL_POLYGON);
        for (i=0, j=0; i < numOfIterations; i++, j += stepSize) {
           glVertex2f(bgCosineTable[j] * radius, bgSineTable[j] * radius);
        }
     glEnd();

     glPopMatrix();
}

// Update the OpenGL display

void bgUpdateDisplay()
{
	// Close the open display list
	glEndList();

	// Update the display
	glutPostRedisplay();
}

/***** Working with Images *****/

// Read in PGM image. Return 1 (TRUE) if successful

#define MAXLINE 80

int bgReadImage (char *filename, int *windW, int *windH, unsigned char **pic)
{
	FILE *fp;
	char line[MAXLINE];
	int i, j, size, rowsize;
	GLubyte *ptr;

	enum ftype {PGM, PPM};	/* possible file types */
	enum ftype filetype;

/* Read in file type */

  fp = fopen(filename, "rb");
  fgets (line, MAXLINE, fp); /* 1st line: PPM or PGM */
  if (line[1] == '5')
	  filetype = PGM;
  else if (line[1] == '6')
	  filetype = PPM;
  else {
	  printf("Error: need PPM or PGM file as input!\n");
	  exit(1);
  }

  /* Read in width and height, & allocate space */

    fgets (line, MAXLINE, fp); /* 2nd line: width height */
    sscanf(line, "%d %d", windW, windH);
    // printf ("Width is %d, height is %d\n", windW, windH);

  if (filetype == PGM) {
    size = *windH * *windW;		/* greymap: 1 byte per pixel */
    rowsize = *windW;
  }
  else /* filetype == PPM */ {
    size = *windH * *windW * 3;	/* pixmap: 3 bytes per pixel */
    rowsize = *windW;
  }
    *pic = (GLubyte *)malloc (size);

    /* Read in maximum value (ignore) */
    fgets (line, MAXLINE, fp); /* 3rd line */

    /* Read in the pixel array row-by-row: 1st row = top scanline */
    ptr = *pic + (*windH-1) * rowsize;
    for (i = *windH; i > 0; i--) {
	  j = fread((void *)ptr, 1, rowsize, fp);
	  ptr -= rowsize;
    }
    fclose(fp);
    return 1;
}

/*
// Take screen image and write out to a PPM file
void bgWriteImage (char *filename)
{
     GLubyte *pixels, *p;
	 pixels = new GLubyte[bgWindW * bgWindH * 3];

     GLubyte rgbval;
     int len, row, col;
     char fname[132];
     FILE *file;
     int width=bgWindW;
     int height=bgWindowDimension-32;
     
//     len = StrLen(filename);
       len = strlen(filename);
     if (len > 128){        // not great, but assuming people aren't specifying
             return;        // such long file names
     }
//     StrCpy(fname,filename);
     strcpy(fname,filename);
//     if (StrCmp(&fname[len-4],".ppm"))
//        StrCat(fname,".ppm");
     if (strcmp(&fname[len-4],".ppm"))
        strcat(fname,".ppm");
     file = fopen(fname, "wb");
     glReadPixels(0, 0, width, height, GL_RGB, 
                     GL_UNSIGNED_BYTE, pixels);
     fprintf(file,"P6\n%d %d\n255\n",width,height);
     for (row=height-1; row>=0; row--)
         for (col=0; col<width; col++)
             for (rgbval=0; rgbval<3; rgbval++)
//               fprintf(file,"%c",pixels[row][col][rgbval]);
                 fwrite(&pixels[row][col][rgbval],sizeof(GLubyte),1,file);
     fclose(file);
}
*/

// Draw image in the window
// use constants to fill window as best as possible
void bgDrawGreyImage (int rows, int columns, unsigned char *image)
{
    float zoom_factor;

    // Start drawing in lower-left corner of the window
	glRasterPos2f(bgLeft, bgBottom);

    // Scale the image to fill the window
	if ((rows / (float)columns) < (bgWindH / (float)bgWindW))
		zoom_factor = (float)bgWindW / columns;
	else
		zoom_factor = (float)bgWindH / rows;
    glPixelZoom (zoom_factor, zoom_factor);

	/* greymap: use as illumination values */
	glDrawPixels(columns, rows, GL_LUMINANCE, GL_UNSIGNED_BYTE, image);
}

void bgDrawHueImage (int rows, int columns, int *image)
{
    float zoom_factor;
    
    // Fix hue values: range 0..180
    for (int i = (rows * columns) - 1; i>= 0; i--)
        image[i] /= 2;
        
    // Start drawing in lower-left corner of the window
	glRasterPos2f(bgLeft, bgBottom);

    // Scale the image to fill the window
    // Scale the image to fill the window
	if ((rows / (float)columns) < (bgWindH / (float)bgWindW))
		zoom_factor = (float)bgWindW / columns;
	else
		zoom_factor = (float)bgWindH / rows;
    glPixelZoom (zoom_factor, zoom_factor);

	/* hue map: use color index */
    glPixelMapfv(GL_PIXEL_MAP_I_TO_R, 256, bgRedHueMap);
    glPixelMapfv(GL_PIXEL_MAP_I_TO_G, 256, bgGreenHueMap);
    glPixelMapfv(GL_PIXEL_MAP_I_TO_B, 256, bgBlueHueMap);
    //glPixelMapfv(GL_PIXEL_MAP_I_TO_A, 256, bgAlphaMap);
	glDrawPixels(columns, rows, GL_COLOR_INDEX, GL_INT, image);
}

void bgDrawColorImage (int rows, int columns, enum BGcolors *image)
{
    float zoom_factor;
    
    // Start drawing in lower-left corner of the window
	glRasterPos2f(bgLeft, bgBottom);

    // Scale the image to fill the window
    // Scale the image to fill the window
	if ((rows / (float)columns) < (bgWindH / (float)bgWindW))
		zoom_factor = (float)bgWindW / columns;
	else
		zoom_factor = (float)bgWindH / rows;
    glPixelZoom (zoom_factor, zoom_factor);

	/* hue map: use color index */
    glPixelMapfv(GL_PIXEL_MAP_I_TO_R, 256, bgRedColorMap);
    glPixelMapfv(GL_PIXEL_MAP_I_TO_G, 256, bgGreenColorMap);
    glPixelMapfv(GL_PIXEL_MAP_I_TO_B, 256, bgBlueColorMap);
    //glPixelMapfv(GL_PIXEL_MAP_I_TO_A, 256, bgAlphaMap);
	glDrawPixels(columns, rows, GL_COLOR_INDEX, GL_INT, (int *)image);
}

void bgDrawPPMatXY (float x, float y, int rows, float width, int columns, unsigned char *image)
{
    float zoom_factor;
    
	// Calculate zoom factor
	zoom_factor = (float)(width * bgWindW) / (float)(rows * bgWidth);
	glPixelZoom (zoom_factor, zoom_factor);

    // Start drawing at specified point
	glRasterPos2f(x, y);

	glDrawPixels(columns, rows, GL_RGB, GL_UNSIGNED_BYTE, (GLubyte *)image);
}

void bgDrawPGMatXY (float x, float y, float width, int rows, int columns, unsigned char *image)
{
    float zoom_factor;
    
	// Calculate zoom factor
	zoom_factor = (float)(width * bgWindW) / (float)(rows * bgWidth);
	glPixelZoom (zoom_factor, zoom_factor);
    
    // Start drawing at specified point
	glRasterPos2f(x, y);

	glDrawPixels(columns, rows, GL_LUMINANCE, GL_UNSIGNED_BYTE, (GLubyte *)image);
}

/***** GLUT functions *****/

// enable GLUT callbacks

void bgEnableCallback(enum BGcallbacks ctype) 
{
	switch (ctype) {
	case BG_KEY: 
		glutKeyboardFunc (bgKeyFunction);
		break;
	case BG_SPECIAL_KEY: 
		glutSpecialFunc (bgSpecialKeyFunction);
		break;
	case BG_IDLE:
		glutIdleFunc (bgIdleFunction);
		break;
	}
}

// disable GLUT callbacks

void bgDisableCallback(enum BGcallbacks ctype) 
{
	switch (ctype) {
	case BG_KEY: 
		glutKeyboardFunc (NULL);
		break;
	case BG_SPECIAL_KEY: 
		glutSpecialFunc (NULL);
		break;
	case BG_IDLE:
		glutIdleFunc (NULL);
		break;
	}
}

// write out text, starting at the given position

void bgWriteString(float x, float y, string s)
{
	int len;
	int c;    // integer version of a character
	int cwid; // width of the character in pixels
	float scale1, scale2; // # of pixels in 1 unit

	scale1 = (float)bgWidth / (bgWindW * 3.0);
	scale2 = (float)(bgWindW * 2.5) / bgWidth;
	glPushMatrix(); // save current matrix

	glScalef(scale1, scale1, 1);
	glTranslatef(x * scale2, y * scale2, 0);	// go to starting position

	len = int(s.size());
	for (int i = 0; i < len; i++) {
		c = (int)s[i];
		glutStrokeCharacter (GLUT_STROKE_ROMAN, c);
		cwid = glutStrokeWidth (GLUT_STROKE_ROMAN, c);
		glTranslatef(cwid*scale1, 0, 0);
	}

	glPopMatrix();  // restore original matrix
}

// Construct the BGimage class with input from a file

BGimage::BGimage (char *filename, bool white_is_transparent)
{
	FILE *fp;
	char line[MAXLINE];
	int i, j, size, rowsize, filerowsize;
	unsigned char *ptr;
	int pold, pnew, pixval, temp;

	enum ftype {PGM, PPM} filetype;	/* possible file types */
	// enum ftype filetype;

/* Read in file type */

  fp = fopen(filename, "rb");
  fgets (line, MAXLINE, fp); /* 1st line: PPM or PGM */
  if (line[1] == '5')
	  filetype = PGM;
  else if (line[1] == '6')
	  filetype = PPM;
  else {
	  printf("Error: need PPM or PGM file as input!\n");
	  exit(1);
  }

  /* Read in width and height, & allocate space */

    fgets (line, MAXLINE, fp); /* 2nd line: width height */
    sscanf(line, "%d %d", &ncols, &nrows);

  if (filetype == PGM) {
	  filerowsize = ncols;
	  if (white_is_transparent) {
		nbytes = 2; 
		type = BG_GREY_ALPHA_IMAGE;
	  }
	  else {
		  nbytes = 1;
		  type = BG_GREY_IMAGE;
	  }
  }
  else /* filetype == PPM */ {
	  filerowsize = ncols * 3;
	  if (white_is_transparent) {
		nbytes = 4; 
		type = BG_RGBA_IMAGE;
	  }
	  else {
		  nbytes = 3;
		  type = BG_RGB_IMAGE;
	  }
  }
	size = nrows * ncols * nbytes;
	rowsize = ncols * nbytes;
    image = new unsigned char[size];

	// printf("Image is %d rows, %d columns, with %d bytes per pixel\n", nrows, ncols, nbytes);

    /* Read in maximum value (ignore) */
    fgets (line, MAXLINE, fp); /* 3rd line */

    /* Read in the pixel array row-by-row: 1st row = top scanline */
    ptr = &(image[0]) + ((nrows-1) * rowsize);
    for (i = nrows; i > 0; i--) {
	  // read in the image
	  j = fread((void *)ptr, 1, filerowsize, fp);
	  ptr -= rowsize;
	  // printf("Read in %d pixels\n", j);
    }
    fclose(fp);

	/* Adjust rows if white is transparent */
	if (type == BG_GREY_ALPHA_IMAGE) {
		/* Grey: every other byte is alpha */
		pnew = size - nbytes; // beginning of last pixel
		for (i = (nrows-1); i >= 0; i--) {
			pold = (i * rowsize) + filerowsize - 1;
			for (j = filerowsize; j > 0; j--, pold --) {
				pixval = image[pold];
				image[pnew] = pixval;
				if (pixval == 255)
					image[pnew + 1] = 0;	// transparent
				else
					image[pnew + 1] = 255;	// opaque
				pnew -=2;
			}
		}
	}
	else if (type == BG_RGBA_IMAGE) {
		/* RGBA: every 4th byte is alpha */
		pnew = size - nbytes; // beginning of last pixel
		for (i = (nrows-1); i >= 0; i--) {
			pold = (i * rowsize) + filerowsize - 3;
			for (j = filerowsize; j > 0; j-=3, pold -=3) {
				pixval = image[pold+2]; image[pnew+2] = pixval;
				temp = image[pold+1];   image[pnew+1] = temp;   pixval += temp;
				temp = image[pold];     image[pnew] = temp;     pixval += temp;
				if (pixval == 765)
					image[pnew+3] = 0;	// transparent
				else
					image[pnew+3] = 255;	// opaque
				pnew -=4;
			}
		}
	}
}

// Construct the BGimage class for given size and type

BGimage::BGimage (int nr, int nc, enum BGimagetypes itype)
{
	// initialize variables
	nrows = nr;
	ncols = nc;
	type = itype;
	if ((type == BG_GREY_IMAGE) || (type == BG_HUE_IMAGE) || (type == BG_COLOR_IMAGE))
		nbytes = 1;
	else if (type == BG_RGB_IMAGE)
		nbytes = 3;
	else if (type == BG_RGBA_IMAGE)
		nbytes = 4;
	else if (type == BG_GREY_ALPHA_IMAGE)
		nbytes = 2;
	image = new unsigned char[nrows * ncols * nbytes];
}

// Draw the BGimage to fill the screen

void BGimage::draw (void)
{
    float zoom_factor;

    // Start drawing in lower-left corner of the window
	glRasterPos2f(bgLeft, bgBottom);

    // Scale the image to fill the window
	if ((nrows / (float)ncols) < (bgWindH / (float)bgWindW))
		zoom_factor = (float)bgWindW / ncols;
	else
		zoom_factor = (float)bgWindH / nrows;
    glPixelZoom (zoom_factor, zoom_factor);

	// Draw the image
	if (type == BG_GREY_IMAGE) 
		glDrawPixels(ncols, nrows, GL_LUMINANCE, GL_UNSIGNED_BYTE, (GLvoid *)image);
	else if (type == BG_HUE_IMAGE) {
		/* hue map: use color index */
		glPixelMapfv(GL_PIXEL_MAP_I_TO_R, 256, bgRedHueMap);
		glPixelMapfv(GL_PIXEL_MAP_I_TO_G, 256, bgGreenHueMap);
		glPixelMapfv(GL_PIXEL_MAP_I_TO_B, 256, bgBlueHueMap);
		glDrawPixels(ncols, nrows, GL_COLOR_INDEX, GL_UNSIGNED_BYTE, (GLvoid *)image);
	}
	else if (type == BG_COLOR_IMAGE) {
		/* color map: use color index */
	    glPixelMapfv(GL_PIXEL_MAP_I_TO_R, 256, bgRedColorMap);
	    glPixelMapfv(GL_PIXEL_MAP_I_TO_G, 256, bgGreenColorMap);
	    glPixelMapfv(GL_PIXEL_MAP_I_TO_B, 256, bgBlueColorMap);
		glDrawPixels(ncols, nrows, GL_COLOR_INDEX, GL_UNSIGNED_BYTE, (GLvoid *)image);
	}
	else if (type == BG_RGB_IMAGE)
		glDrawPixels(ncols, nrows, GL_RGB, GL_UNSIGNED_BYTE, (GLvoid *)image);
	else if (type == BG_GREY_ALPHA_IMAGE) {
		glEnable(GL_BLEND);
		glDrawPixels(ncols, nrows, GL_LUMINANCE_ALPHA, GL_UNSIGNED_BYTE, (GLvoid *)image);
		glDisable(GL_BLEND);
	}
	else if (type == BG_RGBA_IMAGE) {
		glEnable(GL_BLEND);
		glDrawPixels(ncols, nrows, GL_RGBA, GL_UNSIGNED_BYTE, (GLvoid *)image);
		glDisable(GL_BLEND);
	}
}

// Draw the BGimage in specified location, with specified width

void BGimage::draw (float x, float y, float width)
{
    float zoom_factor;
    
	// Calculate zoom factor
	zoom_factor = (float)(width * bgWindW) / (float)(nrows * bgWidth);
	glPixelZoom (zoom_factor, zoom_factor);
    
    // Start drawing at specified point
	glRasterPos2f(x, y);

	if (type == BG_GREY_IMAGE) 
		glDrawPixels(ncols, nrows, GL_LUMINANCE, GL_UNSIGNED_BYTE, (GLvoid *)image);
	else if (type == BG_HUE_IMAGE) {
		/* hue map: use color index */
		glPixelMapfv(GL_PIXEL_MAP_I_TO_R, 256, bgRedHueMap);
		glPixelMapfv(GL_PIXEL_MAP_I_TO_G, 256, bgGreenHueMap);
		glPixelMapfv(GL_PIXEL_MAP_I_TO_B, 256, bgBlueHueMap);
		glDrawPixels(ncols, nrows, GL_COLOR_INDEX, GL_UNSIGNED_BYTE, (GLvoid *)image);
	}
	else if (type == BG_COLOR_IMAGE) {
		/* color map: use color index */
	    glPixelMapfv(GL_PIXEL_MAP_I_TO_R, 256, bgRedColorMap);
	    glPixelMapfv(GL_PIXEL_MAP_I_TO_G, 256, bgGreenColorMap);
	    glPixelMapfv(GL_PIXEL_MAP_I_TO_B, 256, bgBlueColorMap);
		glDrawPixels(ncols, nrows, GL_COLOR_INDEX, GL_UNSIGNED_BYTE, (GLvoid *)image);
	}
	else if (type == BG_RGB_IMAGE)
		glDrawPixels(ncols, nrows, GL_RGB, GL_UNSIGNED_BYTE, (GLvoid *)image);
	else if (type == BG_GREY_ALPHA_IMAGE) {
		glEnable(GL_BLEND);
		glDrawPixels(ncols, nrows, GL_LUMINANCE_ALPHA, GL_UNSIGNED_BYTE, (GLvoid *)image);
		glDisable(GL_BLEND);
	}
	else if (type == BG_RGBA_IMAGE) {
		glEnable(GL_BLEND);
		glDrawPixels(ncols, nrows, GL_RGBA, GL_UNSIGNED_BYTE, (GLvoid *)image);
		glDisable(GL_BLEND);
	}
}




